<?php
/**
 * Whups mail processing/generation library.
 *
 * $Horde: whups/lib/Mail.php,v 1.25 2006/06/26 22:28:39 chuck Exp $
 *
 * Copyright 2004-2006 Jason M. Felice <jfelice@cronosys.com>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 *
 * @author  Jason M. Felice <jfelice@cronosys.com>
 * @package Whups
 */
class Whups_Mail {

    /**
     * Parse a MIME message and create a new ticket.
     *
     * @param string $text  This is the full text of the MIME message.
     * @param array $info   An array of information for the new ticket. This
     *                      should include:
     *                        'queue'    => queue id
     *                        'type'     => type id
     *                        'state'    => state id
     *                        'priority' => priority id
     *                        'ticket'   => ticket id (prevents creation of
     *                                      new tickets)
     * @param string $user  This will be the ticket's requester. If null, we
     *                      will try to deduce from the message's "From"
     *                      field--we do NOT default to Auth::getAuth().
     *
     * @return Whups_Ticket | PEAR_Error  Ticket or Error object.
     */
    function processMail($text, $info, $user = null)
    {
        global $conf, $whups;

        require_once 'Horde/MIME/Structure.php';
        $message = MIME_Structure::parseTextMIMEMessage($text);

        if (preg_match("/^(.*?)\r?\n\r?\n/s", $text, $matches)) {
            $hdrText = $matches[1];
        } else {
            $hdrText = $text;
        }
        $headers = MIME_Structure::parseMIMEHeaders($hdrText, false);

        // If this message was generated by Whups, don't process it.
        if (!empty($headers['X-Whups-Generated'])) {
            return true;
        }

        // Use the message subject as the ticket summary.
        $info['summary'] = Whups_Mail::_findSimpleHeaderValue($headers, 'Subject');
        $info['comment'] = Whups_Mail::_buildComment($message, $headers);

        if (is_null($user)) {
            $user = Whups_Mail::_findAuthUser($headers);
        }
        if (is_null($user) && !empty($info['default_auth'])) {
            $user = $info['default_auth'];
        }

        if (!is_null($user)) {
            if ($user != Auth::getAuth()) {
                $params = Horde::getDriverConfig('auth', $conf['auth']['driver']);
                $auth = &Auth::singleton($conf['auth']['driver'], $params);
                $auth->setAuth($user, array());
            }
            // FIXME: Use Identity to set email address.
        } else {
            if (!empty($conf['mail']['username'])) {
                $params = Horde::getDriverConfig('auth', $conf['auth']['driver']);
                $auth = &Auth::singleton($conf['auth']['driver'], $params);
                $auth->setAuth($conf['mail']['username'], array());
            }
            $from = Whups_Mail::_findSimpleHeaderValue($headers, 'From');
            if (!empty($from)) {
                $info['user_email'] = $from;
            }
        }

        if ($ticket =& Whups_Mail::_findTicket($info)) {
            $ticket->change('comment', $info['comment']);
            return $ticket->commit();
        } elseif (!empty($info['ticket'])) {
            return PEAR::raiseError(sprintf(_("Could not find ticket \"%s\"."),
                                            $info['ticket']));
        } else {
            $result =& Whups_Ticket::newTicket($info, $user);
            return $result;
        }
    }

    function &_findTicket($info)
    {
        if (!empty($info['ticket'])) {
            $ticketnum = $info['ticket'];
        } elseif (preg_match('/\[[\w\s]*#(\d+)\]/', $info['summary'], $matches)) {
            $ticketnum = $matches[1];
        } else {
            $ticket = false;
            return $ticket;
        }

        $ticket = Whups_Ticket::makeTicket($ticketnum);
        if (is_a($ticket, 'PEAR_Error')) {
            $ticket = false;
        }

        return $ticket;
    }

    function _findSimpleHeaderValue($headers, $name)
    {
        foreach ($headers as $hdrName => $hdrVals) {
            if (strcasecmp($hdrName, $name) == 0) {
                if (is_array($hdrVals)) {
                    return implode(', ', $hdrVals);
                } else {
                    return MIME::decode($hdrVals);
                }
            }
        }

        return null;
    }

    function _buildComment(&$message, $headers)
    {
        // Format the message into a comment.
        $comment = "Received message:\n\n";
        foreach ($headers as $name => $vals) {
            if (!strcasecmp($name, 'Subject') || !strcasecmp($name, 'From') ||
                !strcasecmp($name, 'To') || !strcasecmp($name, 'Cc') ||
                !strcasecmp($name, 'Date')) {
                if (is_array($vals)) {
                    foreach ($vals as $val) {
                        $comment .= $name . ': ' . $val . "\n";
                    }
                } else {
                    $comment .= $name . ': ' . MIME::decode($vals) . "\n";
                }
            }
        }

        $comment .= "\n";

        // Look for a text part.
        // FIXME: this is _extremely_ crude.
        $rendered = false;
        if ($message->getType() == 'text/plain') {
            $comment .= $message->transferDecode();
            $rendered = true;
        } else {
            foreach ($message->getParts() as $part) {
                if ($part->getType() == 'text/plain') {
                    $comment .= $part->transferDecode();
                    $rendered = true;
                }
            }
        }
        if (!$rendered) {
            $comment .= "[ Could not render body of message. ]";
        }

        return $comment . "\n";
    }

    /**
     * Search the 'From' header for an email address contained in one
     * of our users' identities.
     *
     * @access private
     */
    function _findAuthUser($headers)
    {
        global $conf;

        require_once 'Horde/Identity.php';
        require_once 'Horde/MIME.php';

        $from = Whups_Mail::_findSimpleHeaderValue($headers, 'From');
        $from = MIME::bareAddress($from);

        $params = Horde::getDriverConfig('auth', $conf['auth']['driver']);
        $auth = &Auth::singleton($conf['auth']['driver'], $params);
        if ($auth->hasCapability('list')) {
            $users = $auth->listUsers();
            foreach ($users as $user) {
                $identity = &Identity::singleton('none', $user);
                $addrs = $identity->getAll('from_addr');
                foreach ($addrs as $addr) {
                    if (strcasecmp($from, $addr) == 0) {
                        return $user;
                    }
                }
            }
        }

        return null;
    }

}
